#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "opcodes.h"
#include "../yad.h"

#define	MAX_ENTRIES	1024
uint16_t table[MAX_ENTRIES];
char *comments[MAX_ENTRIES];
int last = 512;
uint8_t yad_float[64 + 8];
uint8_t yad_3dnow[32];

int arg(uint32_t type)
{
	int a, b, flags, aa = 0;
	
	a = type & 0x0000ff00;	/* op. type */
	b = type & 0x000000ff;	/* op. mode */
	flags = 0;
	if (type & REL)
		flags |= C_REL;
	if (type & STOP)
		flags |= C_STOP;	
	switch (a) {
                case OP_C: case OP_D: case OP_G:
		case OP_P: case OP_PR: case OP_R:
		case OP_S: case OP_V: case OP_VR:

		case OP_M:

		case OP_E: case OP_Q: case OP_W:
			flags |= C_MODRM; aa++;
			break;
		case OP_J:
			flags |= C_REL;
			if (b == b_b)
				flags |= C_ADDR1;
			else
				flags |= C_ADDR4;
			aa++;
			break;
		case OP_O:
			flags |= C_ADDR67; aa++;
			break;
		case OP_A:
			if (b == b_p) {
				flags |= C_DATA66 | C_DATA2 | C_REL | C_STOP;
				aa++;
			}
			break;
		case OP_I:
			switch (b) {
				case b_b:
					flags |= C_DATA1;
					break;
				case b_w:
					flags |= C_DATA2;
					break;
	
				case b_q:
				case b_v:
				case b_z:
					flags |= C_DATA66;
					break;
				default:
					printf("DATA MODE: %02x\n", b);
					exit(2);
			}
			aa++;
			break;
		case OP_F:
		case OP_X:
		case OP_Y:
			aa++;
			break;
		default:
			if (a != 0)
				fprintf(stderr, "Type %08x left unhandled\n", a);
			break;
	}
	return flags;
}

void compile(uint16_t *dst, opcode_t *src, int size, uint16_t tpl_flags)
{
	int i;
	for (i = 0; i < size; i++) {
		if (! strcmp(src[i].name, "(bad)"))
			dst[i] = C_ERROR;
		else
			dst[i] = arg(src[i].op1) | arg(src[i].op2) | arg(src[i].op3) | tpl_flags;
	}
}

char *xstrcat(char *src, char *add) {
	char	*tmp;
	if (src == NULL && add == NULL)
		return (char*)strdup("");
	if (src == NULL || strlen(src) == 0)
		return (char*)strdup(add);
	if (add == NULL || strlen(add) == 0)
		return src;
	tmp = (char*)malloc(strlen(src) + strlen(add) + 1);
	if (tmp == NULL) {
		fprintf(stderr,"OOM!\n");
		exit(2);
	}
	*tmp='\0';
	strcpy(tmp, src);
	free(src);
	strcat(tmp, add);
	return tmp;
}

int add_to_table(uint16_t *what, opcode_t *src, int size, char *rem)
{
	int i, index;
	for (index = -1, i = 0; i < last - size; i++)
		if (! memcmp(&table[i], what, size * 2)) {
			index = i;
			break;
		}
	if (index == -1) {
		index = last;
		for (i = 0; i < size; i++) {
			table[index + i] = what[i];
		}
		last += size;
	}
	for (i = 0; i < size; i++) {
		comments[index + i] = xstrcat(comments[index + i], rem);
		comments[index + i] = xstrcat(comments[index + i], src[i].name);
	}
	return index;
}

void compile_table(void)
{
	int i;
	/* compile regular opcodes */
	for (i = 0; i < 512; i++) {
		if (main_table[i].op1 & (GRP|PRE)) {
			table[i] = 0x55aa;
			continue;
		}
		compile(&table[i], &main_table[i], 1, 0);
		comments[i] = (char*)strdup(main_table[i].name);
	}
	/* compile groups */
	for (i = 0; i < 512; i++) {
		if ((main_table[i].op1 & (GRP|PRE)) == 0)
			continue;
		uint16_t tmp[8], tpl_flags, sz, gflag;
		char rem[16];
		if (i < 256)
			sprintf(rem, " |    %02x ", i);
		else
			sprintf(rem, " | 0f %02x ", i - 256);		
		compile(&tpl_flags, &main_table[i], 1, 0);
		if (main_table[i].op1 & GRP) {
			sz = 8;
			gflag = C_GROUP;
		} else {
			sz = 4;
			gflag = C_PREGRP;
		}
		bzero(tmp, 8 * 2);
		compile(tmp, (opcode_t*)main_table[i].name, sz, tpl_flags);
		table[i] = gflag | add_to_table(tmp, (opcode_t*)main_table[i].name, sz, rem);
	}
}

void compile_float(void)
{
	int i, j, k;
	
	bzero(yad_float, 64 + 8);
	for (i = 0; i < 64; i++)
		if (! strcmp(float_mem[i], "(bad)"))
			yad_float[i / 8] |= (1 << (i % 8));
	for (i = 0; i < 8; i++)
	for (j = 0; j < 8; j++) {
		unsigned char b = 0;
		if (float_reg[i][j].op1 & GRP) {
			for (k = 0; k < 8; k++)
				if (! strcmp(((char**)float_reg[i][j].name)[k], "(bad)"))
					b |= (1 << k);
		} else {
			if (! strcmp(float_reg[i][j].name, "(bad)"))
				b = 0xff;
			else
				b = 0x00;
		}
		yad_float[(i + 1) * 8 + j] = b;
	}
}

void compile_3dnow(void)
{
	int i;
	bzero(yad_3dnow, 32);
	for (i = 0; i < 256; i++)
		if (Cmd3DNow[i] == NULL)
			yad_3dnow[i / 8] |= (1 << (i % 8));
}

void merge_and_dump_tables(void)
{
	int i, j;
	uint8_t	out[2048];
	

#ifdef	PACK6
	uint16_t *values = (uint16_t*)(out + last*3/4);
	int mi = 0, ms = last*3/4;
#else
	uint16_t *values = (uint16_t*)(out + last);
	int mi = 0, ms = last;
#endif
	bzero(out, 2048);
	for (i = 0; i < last; i++) {
		int index = -1;
		for (j = 0; j < mi; j++)
			if (values[j] == table[i]) {
				index = j;
				break;
			}
		if (index == -1) {
			values[mi] = table[i];
			index = mi++;
		}
#ifdef	PACK6
		int n = i*3/4;
		if (index > 63)
			fprintf(stderr, "index too large!\n");
		switch (i & 3) {
			case 0: out[n]	|= index << 2;
				break;
			case 1: out[n]	|= index >> 4;
				out[n+1]|= index << 4;
				break;
			case 2:	out[n]	|= index >> 2;
				out[n+1]|= index << 6;
				break;
			case 3:	out[n]	|= index;
				break;
		}
#else
		out[i] = index;
#endif
	}
	fprintf(stderr, "Number of entries %d\n", last);
	fprintf(stderr, "Number of values %d\n", mi);
	memcpy(&values[mi], yad_float, sizeof(yad_float));
	printf("#define\tYAD_VALUES_OFFSET\t%d\n", ms);
	printf("#define\tYAD_FLOAT_OFFSET\t%d\n", ms + mi * 2);
	ms = ms + mi * 2 + sizeof(yad_float);
	printf("#define\tYAD_3DNOW_OFFSET\t%d\n", ms);
	memcpy(out + ms, yad_3dnow, sizeof(yad_3dnow));
	ms += sizeof(yad_3dnow);
	printf("uint8_t yad_data[%d] = {\n", ms);
	for (i = 0; i < ms; i++) {
		printf("0x%02x%c ", out[i], i == ms - 1 ? ' ' : ',');
		if (i % 16 == 15)
			putchar('\n');
	}
	printf("\n};\n");
}

int main(int argc, char **argv)
{
	compile_table();
	compile_float();
	compile_3dnow();
	merge_and_dump_tables();
	return 0;
}
